import 'package:shared_preferences/shared_preferences.dart';
import 'dart:convert';
import 'package:http/http.dart' as http;
import 'misc.dart';

// data structure for storing leaderboard data
class LeaderboardData {
  // includes an array of json objects sorted in decending point order with the following fields:
  // name, points
  List<Map<String, dynamic>> leaderboardData = [];
  // include a date that the data was last updated
  DateTime lastUpdated = DateTime.now();

  // constructor
  LeaderboardData(this.leaderboardData, this.lastUpdated);

  // constructor to create an empty LeaderboardData object
  LeaderboardData.empty();

  // static method to get the leaderboard data from local storage or the backend if it is not in local storage or if it is more than 1 hour old
  static Future<LeaderboardData> getLeaderboardData(
      [bool invalidateCache = false]) async {
    LeaderboardData? leaderboardData = LeaderboardData.empty();

    // if the cache should be invalidated, get the leaderboard data from the backend
    if (invalidateCache) {
      leaderboardData = await getLeaderboardDataFromBackend();
      // save the leaderboard data to local storage
      await saveLeaderboardDataToLocalStorage(leaderboardData);
      return leaderboardData;
    }
    // get the leaderboard data from local storage
    leaderboardData = await getLeaderboardDataFromLocalStorage();
    // if the leaderboard data is null or if it is more than 1 hour old, get the leaderboard data from the backend
    if (leaderboardData == null ||
        DateTime.now().difference(leaderboardData.lastUpdated).inMinutes >= 1) {
      leaderboardData = await getLeaderboardDataFromBackend();
      // save the leaderboard data to local storage
      await saveLeaderboardDataToLocalStorage(leaderboardData);
    }

    // print(
    // "leaderboardData: ${leaderboardData.leaderboardData}\nlastUpdated: ${leaderboardData.lastUpdated}");
    return leaderboardData;
  }

  Map<String, dynamic> toJson() {
    return {
      'leaderboardData': leaderboardData,
      'lastUpdated': lastUpdated.toIso8601String(),
    };
  }

  // static method to get the leaderboard data from local storage
  static Future<LeaderboardData?> getLeaderboardDataFromLocalStorage() async {
    // get the shared preferences
    SharedPreferences prefs = await SharedPreferences.getInstance();
    // get the leaderboard data from local storage
    var leaderboardDataJson = prefs.getString('leaderboardData');
    // if the leaderboard data is not null, convert it to a LeaderboardData object
    if (leaderboardDataJson != null) {
      return LeaderboardData.fromLocalArrayofJson(leaderboardDataJson);
    }
    return null;
  }

  // static method to get the leaderboard data from the backend
  static Future<LeaderboardData> getLeaderboardDataFromBackend() async {
    // get the leaderboard data from the backend
    var response = await http.get(Uri.parse('$backendUrl/leaderboard'));
    // if the response is successful, convert the response body to a LeaderboardData object
    if (response.statusCode == 200) {
      return LeaderboardData.fromArrayofJson(response.body);
    }
    // if the response is not successful, throw an error
    else {
      throw Exception('Failed to get leaderboard data from backend');
    }
  }

  static Future<void> saveLeaderboardDataToLocalStorage(
      LeaderboardData leaderboardData) async {
    // get the shared preferences
    SharedPreferences prefs = await SharedPreferences.getInstance();
    // save the leaderboard data to local storage
    prefs.setString('leaderboardData', jsonEncode(leaderboardData.toJson()));
  }

  static LeaderboardData fromArrayofJson(String jsonArr) {
    // Decode the JSON array into a dynamic object
    print("jsonArr: $jsonArr");
    List<dynamic> jsonList = jsonDecode(jsonArr);
    // Convert the dynamic object to a List<Map<String, dynamic>>
    List<Map<String, dynamic>> jsonArray =
        jsonList.map((dynamic item) => item as Map<String, dynamic>).toList();

    // Check that the decoded object is a List
    // Map each element of the list to a Map<String, dynamic> object
    return LeaderboardData(jsonArray, DateTime.now());
  }

  static LeaderboardData fromLocalArrayofJson(String jsonArr) {
    // Decode the JSON array into a dynamic object
    Map<String, dynamic> jsonMap = jsonDecode(jsonArr);
    List<dynamic> jsonList = jsonMap['leaderboardData'];
    DateTime lastUpdated = DateTime.parse(jsonMap['lastUpdated']);
    return LeaderboardData(
        jsonList.map((dynamic item) => item as Map<String, dynamic>).toList(),
        lastUpdated);
  }
}
