import 'dart:math';

import 'package:shared_preferences/shared_preferences.dart';
import 'package:http/http.dart' as http;
import 'dart:convert';
import 'misc.dart';
import 'package:flutter/foundation.dart';

class UserData extends ChangeNotifier {
  int currPoints = 0;
  int allTimePoints = 0;
  int allTimeDistance = 0;
  int speedExceededCount = 0;
  int currStreak = 0;
  int allTimeMaxStreak = 0;
  List<num> dailyAvgSpeeds = [];
  List<num> walkDurations = [];
  List<num> walkDistances = [];
  DateTime lastWalkUpdate = DateTime.fromMillisecondsSinceEpoch(0);
  DateTime streakResetDate = DateTime.fromMillisecondsSinceEpoch(0);
  //user reward data
  List<Map<String, dynamic>> inventory = [];

  UserData(
      {this.currPoints = 0,
      this.allTimePoints = 0,
      this.allTimeDistance = 0,
      this.speedExceededCount = 0,
      this.currStreak = 0,
      this.allTimeMaxStreak = 0,
      this.dailyAvgSpeeds = const [],
      this.walkDurations = const [],
      this.walkDistances = const [],
      required this.lastWalkUpdate,
      required this.streakResetDate,
      this.inventory = const []});

  UserData.empty();

  // Converts the UserData object to a JSON string
  String toJson() {
    return jsonEncode({
      'currPoints': currPoints,
      'allTimePoints': allTimePoints,
      'allTimeDistance': allTimeDistance,
      'speedExceededCount': speedExceededCount,
      'currStreak': currStreak,
      'allTimeMaxStreak': allTimeMaxStreak,
      'dailyAvgSpeeds': dailyAvgSpeeds,
      'walkDurations': walkDurations,
      'walkDistances': walkDistances,
      'lastWalkUpdate': lastWalkUpdate.toIso8601String(),
      'streakResetDate': streakResetDate.toIso8601String(),
      'inventory': inventory,
    });
  }

  // Converts a JSON string to a UserData object
  factory UserData.fromJson(String json) {
    Map<String, dynamic> map = jsonDecode(json);

    return UserData(
      currPoints: map['currPoints'] ?? 0,
      allTimePoints: map['allTimePoints'] ?? 0,
      allTimeDistance: map['allTimeDistance'] ?? 0,
      speedExceededCount: map['speedExceededCount'] ?? 0,
      currStreak: map['currStreak'] ?? 0,
      allTimeMaxStreak: map['allTimeMaxStreak'] ?? 0,
      dailyAvgSpeeds: (map['dailyAvgSpeeds'] as List<dynamic>?)
              ?.map((e) => e as num)
              .toList() ??
          [],
      walkDurations: (map['walkDurations'] as List<dynamic>?)
              ?.map((e) => e as num)
              .toList() ??
          [],
      walkDistances: (map['walkDistances'] as List<dynamic>?)
              ?.map((e) => e as num)
              .toList() ??
          [],
      lastWalkUpdate: DateTime.parse(map['lastWalkUpdate'] ??
          DateTime.fromMillisecondsSinceEpoch(0).toString()),
      streakResetDate: DateTime.parse(
          map['streakResetDate'] ?? computeStreakResetDate().toString()),
      inventory: (map['inventory'] as List<dynamic>?)
              ?.map((e) => e as Map<String, dynamic>)
              .toList() ??
          [],
    );
  }

  // Returns a string representation of the UserData object
  @override
  String toString() {
    return 'UserData{currPoints: $currPoints, allTimePoints: $allTimePoints, allTimeDistance: $allTimeDistance, speedExceededCount: $speedExceededCount, currStreak: $currStreak, allTimeMaxStreak: $allTimeMaxStreak, dailyAvgSpeeds: $dailyAvgSpeeds, walkDurations: $walkDurations, walkDistances: $walkDistances, lastWalkUpdate: $lastWalkUpdate, streakResetDate: $streakResetDate, inventory: $inventory}';
  }

  // function that stores this user data locally in SharedPreferences
  Future<void> storeLocally() async {
    SharedPreferences prefs = await SharedPreferences.getInstance();
    prefs.setString('userData', toJson());
  }

  static DateTime computeStreakResetDate() {
    // reset date is the upcoming Sunday at 12:00 AM
    DateTime resetDate = DateTime.now();
    resetDate = resetDate.add(Duration(days: 7 - resetDate.weekday));
    resetDate = DateTime(resetDate.year, resetDate.month, resetDate.day);
    return resetDate;
  }

  bool streakFailed() {
    return DateTime.now().isAfter(
        lastWalkUpdate.add(const Duration(days: 1, hours: 12))); // 36 hours
  }

  void checkStreakReset() {
    if (DateTime.now().isAfter(streakResetDate) || streakFailed()) {
      // streak reset date has passed
      // reset streak
      currStreak = 0;
      // update streak reset date
      streakResetDate = computeStreakResetDate();
    }
  }

  void updateStreak() {
    checkStreakReset();
    currStreak++;
    allTimeMaxStreak = max(allTimeMaxStreak, currStreak);
    lastWalkUpdate = DateTime.now();
  }

  // function that retrieves user data from SharedPreferences
  static Future<UserData> retrieveLocally() async {
    SharedPreferences prefs = await SharedPreferences.getInstance();
    String? userDataJson = prefs.getString('userData');
    if (userDataJson == null) {
      print('WARN: userData not found in SharedPreferences!');
      return UserData(
          currPoints: 0,
          allTimePoints: 0,
          allTimeDistance: 0,
          speedExceededCount: 0,
          currStreak: 0,
          allTimeMaxStreak: 0,
          dailyAvgSpeeds: [],
          walkDurations: [],
          walkDistances: [],
          lastWalkUpdate: DateTime.fromMillisecondsSinceEpoch(0),
          streakResetDate: computeStreakResetDate(),
          inventory: []);
    } else {
      return UserData.fromJson(userDataJson);
    }
  }

  // function that clears user data from SharedPreferences
  static void clearLocally() async {
    SharedPreferences prefs = await SharedPreferences.getInstance();
    prefs.remove('userData');
  }

  // function that updates the local user data with the data from the server
  static Future<bool> updateFromServer() async {
    SharedPreferences prefs = await SharedPreferences.getInstance();
    String? username = prefs.getString('username');
    String? sessionToken = prefs.getString('sessionToken');
    if (username == null) {
      print('WARN: username not found in SharedPreferences!');
      return false;
    }

    if (sessionToken == null) {
      print('WARN: sessionToken not found in SharedPreferences!');
      return false;
    }

    // send a POST request to the server to retrieve the user data
    // body parameters: username, sessionToken
    final response = await http.post(
      Uri.parse('$backendUrl/get-data'),
      headers: {
        'Content-Type': 'application/json',
        'Accept': 'application/json',
      },
      body: json.encode({'username': username, 'sessionToken': sessionToken}),
    );

    if (response.statusCode == 200) {
      // if the request was successful, store the user data locally
      UserData.fromJson(response.body).storeLocally();
      return true;
    } else {
      // if the request was not successful, print the error message
      print('ERROR While Fetching User Data From Server: ${response.body}');
    }

    return false;
  }

  // function that sends the local user data to the server
  static Future<void> sendUserDataToServer() async {
    SharedPreferences prefs = await SharedPreferences.getInstance();
    String? username = prefs.getString('username');
    String? sessionToken = prefs.getString('sessionToken');
    if (username == null) {
      print('WARN: username not found in SharedPreferences!');
      return;
    }

    if (sessionToken == null) {
      print('WARN: sessionToken not found in SharedPreferences!');
      return;
    }

    // retrieve the user data from SharedPreferences
    UserData userData = await UserData.retrieveLocally();

    // send a POST request to the server to update the user data
    // body parameters: username, sessionToken, userData
    final response = await http.post(
      Uri.parse('$backendUrl/set-data'),
      headers: {
        'Content-Type': 'application/json',
      },
      body: json.encode({
        'username': username,
        'sessionToken': sessionToken,
        'data': userData.toJson()
      }),
    );

    if (response.statusCode == 200) {
      // if the request was successful, print the response
      print('INFO: ${response.body}');
    } else {
      // if the request was not successful, print the error message
      print('ERROR While Setting User Data: ${response.body}');
    }

    return;
  }

  void addPoints(int points) {
    currPoints += points;
    allTimePoints += points;
    notifyListeners();
  }

  void updatePointsAfterPurchase() {
    notifyListeners();
  }

  void addStreak(int streak) {
    currStreak += streak;
    allTimeMaxStreak = max(allTimeMaxStreak, currStreak);
  }

  // function that sends a heartbeat to the server
  static Future<void> sendHeartbeat(double latitude, double longitude) async {
    SharedPreferences prefs = await SharedPreferences.getInstance();
    String? sessionToken = prefs.getString('sessionToken');
    if (sessionToken == null) {
      print('WARN: sessionToken not found in SharedPreferences!');
      return;
    }

// send a POST request to the server to send a heartbeat
// body parameters: sessionToken
    final response = await http.post(
      Uri.parse('$backendUrl/walking-heartbeat'),
      headers: {
        'Content-Type': 'application/json',
      },
      body: json.encode({
        'sessionToken': sessionToken,
        'latitude': latitude,
        'longitude': longitude
      }),
    );

    if (response.statusCode == 200) {
      // if the request was successful, print the response
      print('INFO: ${response.body}');
    } else {
      // if the request was not successful, print the error message
      print('ERROR While Sending Heartbeat: ${response.body}');
    }
  }
}
